/* tapplay - TAP file playback for sd2iec hardware
   Copyright (C) 2009  Ingo Korb <ingo@akana.de>

   Inspiration and low-level SD/MMC access based on code from MMC2IEC
     by Lars Pontoppidan et al., see sdcard.c|h and config.h.

   FAT filesystem access based on code from ChaN and Jim Brain, see ff.c|h.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License only.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


   main.c: The interesting stuff

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/crc16.h>
#include <util/delay.h>
#include "config.h"
#include "avrcompat.h"
#include "iec-ll.h"
#include "timer.h"
#include "uart.h"
#include "xmodem.h"

static void (*restart_call)(void) = 0;


static uint8_t databuffer[256];
static volatile uint8_t readidx;
static volatile uint8_t halfwave;
static uint8_t writeidx;
static uint8_t isractive;

static uint8_t longdelay;

ISR(TIMER0_COMPA_vect) {
  static uint16_t delay;
  uint8_t tmp;

  if (!longdelay) {
    tmp = databuffer[readidx];
    toggle_clock();
    toggle_srq();
    if (tmp) {
      OCR0A = (tmp) >> 1;
      if (halfwave) {
        halfwave = 0;
        readidx++;
      } else {
        halfwave = 1;
      }
    } else {
      delay  = databuffer[++readidx] << 8;
      delay |= databuffer[++readidx];
      readidx++;
      longdelay = 1;
      OCR0A = 0x1f;
    }
  } else {
    delay--;
    if (delay == 0) {
      longdelay = 0;
      set_clock(0);
      set_srq(1);
    }
  }
}

static void storeblock(void *buffer, uint32_t fileofs) {
  memcpy(databuffer+writeidx, buffer, 128);
  writeidx += 128;

  if (isractive) {
    // Wait until ISR is in the new half of the buffer
    while ((readidx & 0x80) == (writeidx & 0x80)) ;
  } else {
    // Start send-ISR
    set_data(0);
    readidx = 20; // Skip header
    halfwave = 0;
    set_clock(0);
    set_srq(1);
    _delay_ms(100);
    TIMSK0 |= _BV(OCIE0A);
    isractive = 1;
  }
}

static uint8_t blockbuf[128];

static void init_iec(void) {
#ifdef IEC_SEPARATE_OUT
  /* Set up the port: Output bits as output, all others as input */
  IEC_DDROUT |=            IEC_OBIT_ATN | IEC_OBIT_CLOCK | IEC_OBIT_DATA | IEC_OBIT_SRQ;
  IEC_DDRIN  &= (uint8_t)~(IEC_BIT_ATN  | IEC_BIT_CLOCK  | IEC_BIT_DATA  | IEC_BIT_SRQ);
  /* Enable pullups on the input pins and set the output lines to high */
  IEC_PORT   &= (uint8_t)~(IEC_OBIT_ATN | IEC_OBIT_CLOCK | IEC_OBIT_DATA | IEC_OBIT_SRQ);
  IEC_PORTIN |= IEC_BIT_ATN | IEC_BIT_CLOCK | IEC_BIT_DATA | IEC_BIT_SRQ;
#else
  /* Pullups would be nice, but AVR can't switch from */
  /* low output to hi-z input directly                */
  IEC_DDR  &= (uint8_t)~(IEC_BIT_ATN | IEC_BIT_CLOCK | IEC_BIT_DATA | IEC_BIT_SRQ);
  IEC_PORT &= (uint8_t)~(IEC_BIT_ATN | IEC_BIT_CLOCK | IEC_BIT_DATA | IEC_BIT_SRQ);
#endif
}


int main(void) {
  // JTAG disable
  asm volatile ("in  r24, %0\n"
                "ori r24, 0x80\n"
                "out %0, r24\n"
                "out %0, r24\n"
                :
                : "I" (_SFR_IO_ADDR(MCUCR))
                : "r24"
                );

  init_serial();
  init_timer();
  init_iec();
  sei();

  dprintf("\n\nInitialized, please upload a file with XMODEM-CRC\n");
  uart_flush();

  OCR0A = 0xff;
  TCNT0 = 0;
  TCCR0A = _BV(WGM01);
  TCCR0B = _BV(CS01) | _BV(CS00);

  int32_t len;
  do {
    len = xmodem_receive(blockbuf, storeblock);
    if (len <= 0) {
      dprintf("XMODEM failed: %ld\n",len);
    }
  } while (len <= 0);

  TIMSK0 &= ~_BV(OCIE0A);

  dprintf("Length: %lu\n",len);

  set_data(1);
  set_clock(1);
  set_srq(1);
  _delay_ms(1000);

  cli();
  restart_call();

}
