/* tapplay - TAP file playback for sd2iec hardware
   Copyright (C) 2009  Ingo Korb <ingo@akana.de>


   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License only.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


   xmodem.c: Simple XMODEM-CRC receive-only implementation

*/

#include "config.h"
#include <util/crc16.h>
#include "timer.h"
#include "uart.h"
#include "xmodem.h"

int16_t timed_getc(tick_t maxticks) {
  tick_t starttick = getticks();

  while (time_before(getticks(),starttick+maxticks)) {
    if (uart_gotchar())
      return uart_getc();
  }
  return -1;
}

static void clear_line(void) {
  while (timed_getc(1*HZ) >= 0) ;
}

typedef enum {
  XM_SOH,
  XM_BLOCK,
  XM_INVBLOCK,
  XM_DATA,
  XM_CRCHI,
  XM_CRCLO,
  XM_ERROR
} xmstate_t;

#define SOH 1
#define EOT 4
#define ACK 6
// The specs are unclear on this, sx expects C as NAK
//#define NAK 15
#define NAK 'C'
#define CAN 18

int32_t xmodem_receive(void *buffer, void (*callback)(void *buffer, uint32_t fileofs)) {
  int16_t   c;
  xmstate_t state;
  uint8_t   tmp = 0;
  uint16_t  crc = 0;
  uint8_t   errcount = 0;
  uint8_t   blocknum = 0;
  uint32_t  offset = 0;
  uint8_t   *buf = buffer;

  clear_line();
  do {
    uart_putc(NAK);
    c = timed_getc(5*HZ);
    if (c < 0)
      errcount++;
  } while (errcount < 200 && c != 1);

  if (c != SOH)
    return -1;

  state = XM_BLOCK;
  errcount = 0;

  do {
    /* Try to receive a character */
    c = timed_getc(1*HZ);
    if (c < 0 && state != XM_ERROR) {
      errcount++;
      state = XM_ERROR;
    }

    /* Abort after 10 consecutive errors */
    if (errcount == 10) {
      uart_putc(CAN);
      uart_putc(CAN);
      uart_putc(CAN);
      return -2;
    }

    switch (state) {
    case XM_ERROR:
      /* Drop characters until line is idle */
      clear_line();
      uart_putc(NAK);
      state = XM_SOH;
      break;

    case XM_SOH:
      /* Wait for SOH */
      if (c == SOH) {
        state = XM_BLOCK;
        errcount = 0;
      } else if (c == EOT) {
        /* End of transmission */
        do {
          uart_putc(ACK);
        } while (timed_getc(1*HZ) >= 0);

        return offset + blocknum*128L;
      } else {
        /* Invalid character */
        errcount++;
        state = XM_ERROR;
      }
      break;
      
    case XM_BLOCK:
      /* Block counter */
      tmp = c;
      state = XM_INVBLOCK;
      break;

    case XM_INVBLOCK:
      /* Inverted block counter, should match either
       * the current or previous block
       */
      if (c == (tmp ^ 0xff) &&
          (tmp == blocknum || tmp == ((blocknum+1) & 0xff))) {
        blocknum = tmp;
        state = XM_DATA;
        tmp = 0;
        crc = 0;
      } else {
        state = XM_ERROR;
      }
      break;

    case XM_DATA:
      /* Data */
      buf[tmp++] = c;
      crc = _crc_xmodem_update(crc, c);
      if (tmp == 128)
        state = XM_CRCHI;
      break;

    case XM_CRCHI:
      /* High byte of the CRC */
      tmp = c;
      state = XM_CRCLO;
      break;

    case XM_CRCLO:
      /* Low byte of the CRC */
      if (crc == ((tmp << 8) | c)) {
        if (blocknum == 0)
          offset += 32768L;
        callback(buffer, offset + 128*(blocknum-1L));
        uart_putc(ACK);
        state = XM_SOH;
      } else {
        uart_putc(NAK);
        state = XM_SOH;
      }
      break;
    }
  } while (1);
}
